"""
Test PsqlWrapper, and secondarily PsqlParser.
"""

from ..psqlwrapper import PsqlWrapper
from ..psqlparser import PsqlParser

from shutil import get_terminal_size


def new_psqlwrapper() -> PsqlWrapper:
    psql = PsqlWrapper("",
                       lambda x: "Test",
                       lambda x: "",
                       PsqlParser())

    case_psql_start = \
        b'psql (14.5)\r\nType "help" for help.\r\n\r\n\x1b[?2004hpgdb=# '
    psql._intercept(case_psql_start)
    return psql

def test_intercept() -> None:
    psql: PsqlWrapper = new_psqlwrapper()

    fresh_prompt_1 = \
        b'\x1b[?2004hpgdb=# '
    return_press_1 = \
        b'\r\n'
    return_press_2 = \
        b'\r\n\x1b[?2004l\r'
    return_press_3 = \
        b'\x08\r\n'

    # simple query with single arrow-up to query
    case_query_1 = \
        b'SELECT * FROM orders WHERE order_total_eur = 0 AND order_total_eur = 100;'
    psql._intercept(case_query_1)
    psql._intercept(return_press_1)
    assert psql.pg4n_message == "Test"
    
    assert psql._intercept(fresh_prompt_1) == \
        b'\r\n' + b'Test' + b'\r\n\r\n' + b'\x1b[?2004hpgdb=# '

    psql = new_psqlwrapper()

    # Something weird going these test case. Pyte understands this case right
    # in live scenarios most of the time, but I cannot reproduce it with this
    # saved stream of characters. Most likely the control streams depend
    # on terminal size, and cannot be just copy-pasted from live tests.
    
    # ctrl-R to previous query with three key presses searching (1..0..0). Press return.
    case_query_2 = \
        b"\x1b[A\rtest_db=# SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur = 100;\x1b[K\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\r\n\r"
    psql._intercept(b"'\r(reverse-i-search)`': ")
    psql._intercept(b"\x08\x08\x08=': SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur \x1b[7m=\x1b[27m 100;\x08\x08\x08\x08\x08\x08")
    psql._intercept(b"\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C1': SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur \x1b[7m= 1\x1b[27m00;\x08\x08\x08\x08\x08\x08")
    psql._intercept(b"\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C0': SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur \x1b[7m= 10\x1b[27m0;\x08\x08\x08\x08\x08\x08")
    psql._intercept(b"\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C0': SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur \x1b[7m= 100\x1b[27m;\x08\x08\x08\x08\x08\x08")
    psql._intercept(b"\x1b[A\rtest_db=# SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur = 100;\x1b[K\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\r\n\r")
    psql._intercept(b"\r\n\x1b[?2004l\r")
    assert psql.pg4n_message == "Test"
    case_query_2_prompt = b" order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004htest_db=# "
    assert psql._intercept(case_query_2_prompt) == \
        b" order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\r\nTest\r\n\r\n\x1b[?2004htest_db=# "
    psql = new_psqlwrapper()

    # Arrow-up to previous query, alt-B until 'orders   WHERE' and remove the
    # extra 2 whitespaces:
    psql._intercept(b'SELECT * FROM orders WHERE   order_total_eur = 0 AND order_total_eur = 100;')
    psql._intercept(b'\r\x1b[C')
    psql._intercept(b'\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C')
    psql._intercept(b'\x08\x08\x08\x08\x08\x08')
    psql._intercept(b'\x08\x08\x08\x08\x08\x08')
    psql._intercept(b'\x08\x08\x08\x08')
    psql._intercept(b'\x08\x08')
    psql._intercept(b'\x08\x08\x08\x08\x08\x08')
    psql._intercept(b'\x08\x08\x08\x08\x08\x08')
    psql._intercept(b'\x08\x08\x08\x08\x08\x08')
    psql._intercept(b'\x08order_total_eur = 0 AND order_total_eur = 1\x1b[1P00;\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C')
    psql._intercept(b'\x08order_total_eur = 0 AND order_total_eur = 10\x1b[C\x1b[1P;\x1b[A\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C\x1b[C')
    psql._intercept(b'\r\n\r\r\n')
    assert psql.pg4n_message == "Test"
    
    assert psql._intercept(b'\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\x1b[?2004htest_db=# ') == \
        b'\x1b[?2004l\r order_id | order_total_eur | customer_id \r\n----------+-----------------+-------------\r\n(0 rows)\r\n\r\n\r\nTest\r\n\r\n\x1b[?2004htest_db=# '

    psql = new_psqlwrapper()

    # multiline query
    psql._intercept(b'SELECT')
    psql._intercept(b'\r\n\x1b[?2004l\r')
    psql._intercept(b'*')
    psql._intercept(b'\r\n\x1b[?2004l\r')
    psql._intercept(b'\x1b[?2004htest_db-# ')
    psql._intercept(b'FROM')
    psql._intercept(b'\r\n')
    psql._intercept(b'\x1b[?2004l\r\x1b[?2004htest_db-# ')
    psql._intercept(b'orders')
    psql._intercept(b'\r\n')
    psql._intercept(b'\x1b[?2004l\r\x1b[?2004htest_db-# ')
    psql._intercept(b';')
    psql._intercept(b'\r\n')
    assert psql.pg4n_message == "Test"
    psql._intercept(b'\x1b[?2004l\r\x1b[?1049h\x1b=\r order_id | order_total_eur | customer_id \x1b[m\r\n----------+-----------------+-------------\x1b[m\r\n        1 |          535.36 |         111\x1b[m\r\n        2 |          409.80 |         217\x1b[m\r\n        3 |          189.43 |          19\x1b[m\r\n        4 |          144.14 |         157\x1b[m\r\n        5 |          582.52 |         172\x1b[m\r\n        6 |          132.85 |         206\x1b[m\r\n        7 |          183.92 |         236\x1b[m\r\n        8 |          424.80 |         244\x1b[m\r\n        9 |          519.43 |         175\x1b[m\r\n       10 |          414.55 |         234\x1b[m\r\n       11 |           88.19 |          50\x1b[m\r\n       12 |          591.72 |         143\x1b[m\r\n       13 |          503.52 |         216\x1b[m\r\n       14 |          586.06 |         181\x1b[m\r\n       15 |           47.79 |         248\x1b[m\r\n       16 |          330.92 |         130\x1b[m\r\n       17 |          302.31 |         225\x1b[m\r\n       18 |          438.38 |          26\x1b[m\r\n       19 |          107.53 |          94\x1b[m\r\n       20 | '
b'         207.60 |           9\x1b[m\r\n       21 |          471.12 |         179\x1b[m\r\n:\x1b[K')
    psql._intercept(b'\r\x1b[K\x1b>\x1b[r\x1b[?1049l')
    assert psql._intercept(b'\x1b[?2004htest_db=# ') == \
        b'\r\nTest\r\n\r\n\x1b[?2004htest_db=# '
    



import os.path
import sys
from os import getcwd, getenv
from typing import Optional

from .config_parser import ConfigParser
from .config_values import ConfigValues

CONFIG_FILE_NAME = "pg4n.conf"


class ConfigReader:
    def __init__(self):
        pass

    def read(self) -> Optional[ConfigValues]:
        """
        The configuration files are read in order from: /etc/pg4n.conf then from
        $XDG_CONFIG_HOME/pg4n.conf, or if $XDG_CONFIG_HOME is not set, from
        $HOME/.config/pg4n.conf, and lastly from $PWD/pg4n.conf, with each new
        value introduced in latter files overriding the previous value.
        """

        config_fnames = []
        config_values: ConfigValues = {}

        xdg_config_home = getenv("XDG_CONFIG_HOME")
        home = getenv("HOME")

        etc_config_path = "/etc/" + CONFIG_FILE_NAME
        if os.path.isfile(etc_config_path):
            config_fnames.append(etc_config_path)

        if xdg_config_home:
            custom_config_home_path = xdg_config_home + "/" + CONFIG_FILE_NAME
            if os.path.isfile(custom_config_home_path):
                config_fnames.append(custom_config_home_path)
        else:
            home_config_path = home + "/.config/" + CONFIG_FILE_NAME
            if os.path.isfile(home_config_path):
                config_fnames.append(home_config_path)

        # As a Last resort, we check the current directory for a config file.
        try:
            cwd = os.getcwd()
            cwd_config_path = cwd + "/" + CONFIG_FILE_NAME
            if os.path.isfile(cwd_config_path):
                config_fnames.append(cwd_config_path)
        except Exception as e:
            print(
                f"error: unable to get current working directory '{e}'", file=sys.stderr
            )

        if len(config_fnames) == 0:
            return None

        for config_fname in config_fnames:
            try:
                with open(config_fname, "r") as config_file:
                    config_parser = ConfigParser(config_file)
                    new_config_values: ConfigValues = config_parser.parse()
                    if new_config_values is not None:
                        for k, v in new_config_values.items():
                            config_values[k] = v
            except Exception as e:
                if hasattr(e, "errno"):
                    print(
                        f"error: unable to read config file: '{config_fname}' [str({e.errno})]",
                        file=sys.stderr,
                    )
                else:
                    print(
                        f"error: unable to read config file: '{config_fname}'",
                        file=sys.stderr,
                    )

        return config_values



from ..errfmt import ErrorFormatter

VT100_UNDERLINE = "\x1b[4m"
VT100_RESET = "\x1b[0m"


def test_format():
    warning = "Too many parentheses"
    warning_name = "SurplusParentheses"

    formatter = ErrorFormatter(warning, warning_name)
    warning_msg = formatter.format()
    assert warning_msg.find(warning) != -1 and warning_msg.find(warning_name) != -1

    warning = "Too many semicolons"
    warning_name = "SurplusSemicolons"
    underlined_query = f"""SELECT *
FROM customers
{VT100_UNDERLINE}WHERE type 'C'{VT100_RESET} = 1 OR 100 = 100;"""

    formatter = ErrorFormatter(warning, warning_name, underlined_query)
    warning_msg = formatter.format()
    assert (
        warning_msg.find(warning) != -1
        and warning_msg.find(warning_name) != -1
        and warning_msg.find(underlined_query) != -1
    )



import re
from dataclasses import dataclass
from typing import Optional

import sqlglot.expressions as exp

from .sqlparser import SqlParser
from .sqlparser import Column
from .errfmt import ErrorFormatter


VT100_UNDERLINE = "\x1b[4m"
VT100_RESET = "\x1b[0m"


@dataclass(frozen=True)
class CmpContext:
    expression: exp.Predicate
    a: Column
    b: Column


class CmpDomainChecker:
    def __init__(self, parsed_sql: exp.Expression, columns: list[Column]):
        self.parsed_sql: str = parsed_sql
        self.columns: list[Column] = columns
        self.suspicious_cmp_contexts: list[CmpContext] = []
        self.warning_msg: Optional[str] = None

    def check(self) -> Optional[str]:
        """
        Does analysis for suspicous comparisons between different domains.
        e.g., comparing columns off type VARCHAR(20) and VARCHAR(50)
        Returns a warning message if something was found, otherwise None.
        """

        self._detect_suspicious_cmps(self.parsed_sql, self.columns)
        return self.warning_msg

    def _are_from_compatible_domains(self, a: Column, b: Column) -> bool:
        """
        Tests whether the both types a and b are representable by their
        datatype.
        """

        # First we must extract the base type without possible trailing
        # precission or digit counts (which can be retrieved from the
        # Column.type directly).
        type_start_matcher = re.compile(r'^([a-zA-Z]+).*$')
        match_a = type_start_matcher.match(a.type.name)
        match_b = type_start_matcher.match(b.type.name)

        if not (match_a and match_b):
            return True
        elif len(match_a.groups()) != 1 or len(match_b.groups()) != 1:
            return True

        type_start_a = match_a.group(1)
        type_start_b = match_b.group(1)

        # We don't evaluate types of different names
        # TODO: Consider evaluating similar types e.g. VARCHAR and CHAR
        if type_start_a != type_start_b:
            return True

        if a.type.digits is None and b.type.digits is None:
            return True
        elif ((a.type.digits is None and b.type.digits is not None) or
              (a.type.digits is not None and b.type.digits is None)):
            return False

        if a.type.digits != b.type.digits:
            return False
        else:
            if a.type.precision is None and b.type.precision is None:
                return True
            elif ((a.type.precision is None and
                   b.type.precision is not None) or
                  (a.type.precision is not None and
                   b.type.precision is None)):
                return False

        return True

    def _detect_suspicious_cmp(self, cmp: exp.Predicate,
                               columns: list[Column]):
        """
        Detects whether 'cmp' has comparison between columns from different
        domains (e.g. a: VARCHAR(10) < b: VARCHAR(50)).

        Only works for comparisons between 2 variables. In other words, if
        any operand in the comparison is literal this function returns False.

        Ignores any casts used for the operands.
        """

        # TODO: Take casts into account.
        # TODO: Consider comparisons with literals as well.
        # TODO: Investigate if postgresql already errors/warns of these kind of
        #       comparisons
        # FIXME: Stop assuming columns with the same name coming from possibly
        #        different tables have the same datatype
        column_expressions = cmp.find_all(exp.Column)

        cmp_column_names = []
        cmp_columns = []

        for column_expression in column_expressions:
            cmp_column_name = \
                SqlParser.get_column_name_from_column_expression(
                    column_expression)
            cmp_column_names.append(cmp_column_name)

        for cmp_column_name in cmp_column_names:
            tmp = _find_column(cmp_column_name, columns)
            if tmp is not None:
                cmp_columns.append(tmp)

        # This the comparisons has atleast 1 literal
        if len(cmp_columns) < 2:
            return False

        if not self._are_from_compatible_domains(cmp_columns[0],
                                                 cmp_columns[1]):
            cmp_context = CmpContext(cmp, cmp_columns[0], cmp_columns[1])
            self.suspicious_cmp_contexts.append(cmp_context)

    def _detect_suspicious_cmps(self, select_statement: exp.Select,
                                columns: list[Column]):
        predicates = SqlParser.find_where_predicates(select_statement)

        # This filters predicates we are not interested in such as IN or EXISTS
        binary_predicates = list(
            filter(lambda x: isinstance(x, exp.Binary), predicates))

        for binary_predicate in binary_predicates:
            self._detect_suspicious_cmp(binary_predicate, columns)

        if len(self.suspicious_cmp_contexts) == 0:
            return

        for i, suspicious_cmp_context in \
                enumerate(self.suspicious_cmp_contexts):
            if self.warning_msg is None:
                self.warning_msg = ""
            whole_statement = str(select_statement)

            domain1 = suspicious_cmp_context.a.type.name
            domain2 = suspicious_cmp_context.b.type.name
            warning = f"Comparison between different domains ({domain1}, {domain2})"
            # TODO: Develop more principled way of matching warning_name with
            #       the options expected in the configuration files.
            warning_name = type(self).__name__.rstrip("Checker")

            cmp_exp = suspicious_cmp_context.expression

            # It does not matter which column's ancestor Where expression we
            # find because both necessarily have the same.
            containing_where = str(cmp_exp.find_ancestor(exp.Where))
            containing_where_start_offset = \
                whole_statement.find(containing_where)

            cmp_start_offset = containing_where.find(str(cmp_exp))
            cmp_end_offset = cmp_start_offset + len(str(cmp_exp))

            total_start_offset = \
                containing_where_start_offset + cmp_start_offset
            total_end_offset = containing_where_start_offset + cmp_end_offset

            underlined_query = \
                whole_statement[:total_start_offset] + \
                VT100_UNDERLINE + \
                whole_statement[total_start_offset:total_end_offset] + \
                VT100_RESET + \
                whole_statement[total_end_offset:len(whole_statement)]

            formatter = ErrorFormatter(warning, warning_name, underlined_query)
            self.warning_msg = formatter.format()
            if i != len(self.suspicious_cmp_contexts) - 1:
                self.warning_msg += '\n'


def _find_column(column_name: str, columns: list[Column]) -> Optional[Column]:
    """
    Returns a column matching column name.
    """
    # TODO: Just use dict inside an object

    for column in columns:
        if column_name == column.name:
            return column

    return None



from typing import Optional
import copy

from psycopg import Connection
import sqlglot.expressions as exp
from .sqlparser import (
    SqlParser,
    Column,
)
from .qepparser import (
    QEPParser,
    QEPAnalysis,
    QEPNode,
)
from .errfmt import ErrorFormatter


class ImpliedExpressionChecker:
    def __init__(self, parsed_sql: exp.Expression, sql_statement: str,
                 db_connection: Connection):
        self.parsed_sql: exp.Expression = parsed_sql
        self.sql_statement: str = sql_statement
        self.db_connection: Connection = db_connection

    def check(self) -> Optional[str]:
        """
        Returns warning_msg if implied expression is detected, otherwise None.
        """

        def finder(node: QEPNode) -> bool:
            return node.get("One-Time Filter") != None

        qep_parser_with_constraint_exclusion = \
            QEPParser(conn=self.db_connection, constraint_exclusion=True)
        qep_analysis_with_constraint_exclusion = \
            qep_parser_with_constraint_exclusion.parse(self.sql_statement)
        if qep_analysis_with_constraint_exclusion is None:
            return None
        has_onetime_filter_with_constraint_exclusion = \
            len(qep_analysis_with_constraint_exclusion.root.rfind(finder)) > 0

        qep_parser_without_constraint_exclusion = \
            QEPParser(conn=self.db_connection, constraint_exclusion=False)
        qep_analysis_without_constraint_exclusion = \
            qep_parser_without_constraint_exclusion.parse(self.sql_statement)
        if qep_analysis_without_constraint_exclusion is None:
            return None
        has_onetime_filter_without_constraint_exclusion = \
            len(qep_analysis_without_constraint_exclusion.root.rfind(finder)) > 0

        has_implied_expression = \
            has_onetime_filter_with_constraint_exclusion and \
            not has_onetime_filter_without_constraint_exclusion

        if not has_implied_expression:
            return None

        warning = "Found impossible comparison due to column/table constraints"
        warning_name = type(self).__name__.rstrip("Checker")

        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg



from typing import Optional

import sqlglot.expressions as exp

from .qepparser import QEPAnalysis
from .errfmt import ErrorFormatter


class SumDistinctChecker:
    def __init__(self, parsed_sql: exp.Expression, qep_analysis: QEPAnalysis):
        self.parsed_sql = parsed_sql
        self.qep_analysis = qep_analysis

    def check(self) -> Optional[str]:
        """
        Returns warning message if the sql has SUM/AVG(DISTINCT ...),
        otherwise None
        """

        sums = self.parsed_sql.find_all(exp.Sum)
        avgs = self.parsed_sql.find_all(exp.Avg)

        def has_distinct(x):
            return type(x.this) == exp.Distinct

        has_sum_distinct = any(filter(has_distinct, sums))
        has_avg_distinct = any(filter(has_distinct, avgs))

        if not (has_sum_distinct or has_avg_distinct):
            return None

        warning = "DISTINCT in SUM or AVG"
        warning_name = type(self).__name__.rstrip("Checker")
        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pg4n', 'pg4n.test']

package_data = \
{'': ['*']}

install_requires = \
['myst-parser>=0.18.1,<0.19.0',
 'pexpect>=4.8.0,<5.0.0',
 'psycopg[binary]>=3.1.3,<4.0.0',
 'pyparsing>=3.0.9,<4.0.0',
 'pyte>=0.8.1,<0.9.0',
 'pytest-postgresql>=4.1.1,<5.0.0',
 'pytest>=7.1.3,<8.0.0',
 'sphinx-autodoc-typehints>=1.19.4,<2.0.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0',
 'sphinx>=5.2.3,<6.0.0',
 'sqlglot>=6.3.2,<7.0.0']

entry_points = \
{'console_scripts': ['pg4n = pg4n:main.main']}

setup_kwargs = {
    'name': 'pg4n',
    'version': '1.0.0.56',
    'description': 'PostgreSQL for novices - helps debug PostgreSQL queries',
    'long_description': "# PostgreSQL for novices\n\n[📄 Documentation](https://project-c-sql.github.io/)\n\nThis README is meant for developers of the project, and not for end users. For end users, please see the documentation linked above.\n\n- [PostgreSQL for novices](#postgresql-for-novices)\n  - [Notes for developers](#notes-for-developers)\n    - [Poetry](#poetry)\n      - [Versioning](#versioning)\n    - [Imports](#imports)\n    - [Running tests](#running-tests)\n      - [Using docker](#using-docker)\n    - [Building documents](#building-documents)\n    - [Linting](#linting)\n\n## Notes for developers\n\n### Poetry\n\nThis project uses [Poetry](https://python-poetry.org/) for packaging. Although one should refer to [Poetry docs](https://python-poetry.org/docs/) for a thorough introduction, here's a short summary of the intended workflow with Poetry:\n\n- To install all dependencies and the application, type `poetry install`. After installation, if the Python scripts folder is in your PATH, you should be able to invoke `main.main()` with `pg4n`.\n- To make VS Code use Poetry's virtual environment, type `poetry env info`, copy virtual environment executable path, press F1 and type `Python: Select Interpreter` > `Enter interpreter path...` > paste path and press `<ENTER>`.\n- To add/remove a dependency, type `poetry add <dep>`/`poetry remove <dep>`.\n- To execute a command from within virtual environment shell, type `poetry run <cmd>`.\n- To enter a shell session within the Poetry virtual environment, type `poetry shell`.\n\n#### Versioning\n\nYou can bump the version number automatically with `poetry version patch`, `poetry version minor`, etc. See `poetry version -h`.\n\nSee version history [here](https://pypi.org/project/pg4n/#history).\n\n### Imports\n\nDuring development, you must run the program as a module, e.g., `poetry run python -m src.pg4n.main`, so that the imports work.\n\n### Running tests\n\nHaving PostgreSQL running on port 5432, do `poetry run pytest`.\n\nYou may need to provide environment variables that match your config:\n\n| Variable     | Default value   | Description                                             |\n| ------------ | --------------- | ------------------------------------------------------- |\n| `PGHOST`     | `127.0.0.1`     | Hostname of the PostgreSQL server.                      |\n| `PGPORT`     | `5432`          | Port to an active PostgreSQL instance.                  |\n| `PGUSER`     | `postgres`      | The user that will be used to manage the test database. |\n| `PGPASSWORD` |                 | Password, in case password authentication is used.      |\n| `PGDBNAME`   | `test_database` | Database name.                                          |\n \nFor example, if PostgreSQL is on port 5433, just do `PGPORT=5433 poetry run pytest` (Bash syntax).\n\n#### Using docker\n\nTo get a similar PostgreSQL instance as with GitHub Actions workflow:<br>\n`docker run --rm -P -p 127.0.0.1:5432:5432 --name pg -e POSTGRES_PASSWORD=postgres -d postgres:14.5-alpine`\n\nYou'll need to tell pytest the password: `PGPASSWORD=postgres poetry run pytest`.\n\n### Building documents\n\n1. If `docs/api` is not up-to-date or doesn't exist, run:<br>`poetry run sphinx-apidoc -f -o docs/api src/pg4n '*/test*'`\n2. To generate the documentation:<br>`poetry run sphinx-build -b html docs docs/build`\n\nNote that the GitHub Pages site is only updated on pushes to `main` branch.\n\n### Linting\n\nFor linting, you need the CI tools: `poetry install --with=ci`. The tools used are:\n- `black` for formatting\n- `pylint` for linting\n- `mypy` for static type checking\n- `isort` for sorting imports\n\nTo get a grade that the CI/CD pipeline would give you, you can do `poetry run scripts/ci-grade.sh` to run all the checks. The output is possibly long, so pipe it to a file perusal filter such as `less` to scroll through it and search for things of concern, e.g., `summary` to see scores.\n",
    'author': 'Joni Nikki',
    'author_email': 'joni.nikki@tuni.fi',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from typing import Optional

import sqlglot.expressions as exp

from .sqlparser import SqlParser
from .errfmt import ErrorFormatter

VT100_UNDERLINE = "\x1b[4m"
VT100_RESET = "\x1b[0m"


@dataclass(frozen=True)
class SubquerySelectContext:
    subquery: exp.Expression


class SubquerySelectChecker:
    def __init__(self, parsed_sql: exp.Expression, sql_parser: SqlParser):
        self.parsed_sql: exp.Expression = parsed_sql
        self.sql_parser: SqlParser = sql_parser
        self.nested_condition_contexts: list[SubquerySelectContext] = []

    def check(self) -> Optional[str]:
        """
        Returns warning message if there no column SELECTed in a subquery is
        not used in that subquery of its own columns, otherwise returns None.
        """
        self._detect_suspicious_nested_conditions()

        if len(self.nested_condition_contexts) == 0:
            return None

        warning_msg = ""

        warning = "No column in subquery SELECT references its tables"
        warning_name = type(self).__name__.rstrip("Checker")


        for i, nested_condition_context in enumerate(self.nested_condition_contexts):
            whole_statement = str(self.parsed_sql)
            subquery = str(nested_condition_context.subquery)
            subquery_start_offset = whole_statement.find(subquery)
            subquery_end_offset = subquery_start_offset + len(subquery)

            underlined_query = (
                whole_statement[:subquery_start_offset]
                + VT100_UNDERLINE
                + subquery
                + VT100_RESET
                + whole_statement[subquery_end_offset : len(whole_statement)]
            )

            formatter = ErrorFormatter(warning, warning_name, underlined_query)
            warning_msg += formatter.format()

            if i != len(self.nested_condition_contexts) - 1:
                warning_msg += "\n"

        return warning_msg

    def _detect_suspicious_nested_conditions(self):
        # exp.In is not SubqueryPredicate for some reason
        subquery_predicates = self.parsed_sql.find_all(exp.SubqueryPredicate)
        in_expressions = self.parsed_sql.find_all(exp.In)
        in_subqueries = [x.args.get("query") for x in in_expressions]
        subqueries = list(subquery_predicates) + in_subqueries

        if subqueries is None:
            return

        # We need to find whether the subquery SELECT uses a tuple variable
        # (e.g. FROM statement) of the subquery.
        for subquery in subqueries:
            all_subquery_columns = self.sql_parser.get_query_columns(subquery)
            all_subquery_column_names = [x.name for x in all_subquery_columns]

            select_column_names = []
            select = subquery.find(exp.Select)
            for select_expression in select.expressions:
                column_exps = select_expression.find_all(exp.Column)
                for column_exp in column_exps:
                    column_name = SqlParser.get_column_name_from_column_expression(
                        column_exp
                    )
                    select_column_names.append(column_name)

            if not any(
                filter(lambda x: x in all_subquery_column_names, select_column_names)
            ):
                context = SubquerySelectContext(subquery)
                self.nested_condition_contexts.append(context)



import pytest
import psycopg

from os import getenv
from pytest_postgresql import factories
from psycopg import Connection

from .. import qepparser


def load_database(**kwargs):
    conn: Connection = psycopg.connect(**kwargs)
    with conn.cursor() as cur:
        cur.execute("""
        -- for copy-and-pasting
        drop table if exists comments;
        drop table if exists users;
        drop table if exists stories;

        -- demonstrates relational data
        create table stories (id serial primary key, name varchar);
        create table users (id serial primary key, name varchar);
        create table comments (
            id serial primary key,
            story_id integer references stories(id) on delete cascade,
            user_id integer references users(id) on delete cascade,
            comment varchar);

        -- populate with sample data
        insert into stories (name) values ('story1');
        insert into stories (name) values ('story2');
        insert into users (name) values ('user1');
        insert into users (name) values ('user2');
        insert into comments (story_id, user_id, comment) values (1, 1, 'comment1');
        insert into comments (story_id, user_id, comment) values (1, 2, 'comment2');
        insert into comments (story_id, user_id, comment) values (2, 1, 'comment3');
        insert into comments (story_id, user_id, comment) values (2, 2, 'comment4');
        """)
        conn.commit()


postgresql_in_docker = factories.postgresql_noproc(
    load=[load_database],
    host=getenv("PGHOST", "127.0.0.1"),
    port=getenv("PGPORT", 5432),
    user=getenv("PGUSER", "postgres"),
    password=getenv("PGPASSWORD"),
    dbname=getenv("PGDBNAME", "test_database"))
postgresql = factories.postgresql("postgresql_in_docker")


@pytest.fixture
def parser(postgresql: Connection):
    return qepparser.QEPParser(conn=postgresql)


def test_qep_structure(parser: qepparser.QEPParser):
    """Test that the QEP structure is as expected."""

    qep = parser("select * from stories")
    assert qep.plan["Node Type"] == "Seq Scan"
    assert qep.plan["Relation Name"] == "stories"
    assert qep.plan["Alias"] == "stories"
    assert qep.plan["Actual Rows"] == 2

    qep = parser("select * from users")
    assert qep.plan["Node Type"] == "Seq Scan"
    assert qep.plan["Relation Name"] == "users"
    assert qep.plan["Alias"] == "users"
    assert qep.plan["Actual Rows"] == 2

    qep = parser("select * from comments")
    assert qep.plan["Node Type"] == "Seq Scan"
    assert qep.plan["Relation Name"] == "comments"
    assert qep.plan["Alias"] == "comments"
    assert qep.plan["Actual Rows"] == 4

    qep = parser("select * from stories where id = 1")
    assert qep.plan["Node Type"] == "Index Scan"
    assert qep.plan["Relation Name"] == "stories"
    assert qep.plan["Alias"] == "stories"
    assert qep.plan["Actual Rows"] == 1

    qep = parser("select * from users where id = 1")
    assert qep.plan["Node Type"] == "Index Scan"
    assert qep.plan["Relation Name"] == "users"
    assert qep.plan["Alias"] == "users"
    assert qep.plan["Actual Rows"] == 1

    qep = parser("select * from comments where id = 1")
    assert qep.plan["Node Type"] == "Index Scan"
    assert qep.plan["Relation Name"] == "comments"
    assert qep.plan["Alias"] == "comments"
    assert qep.plan["Actual Rows"] == 1

    qep = parser("select * from stories where id = 1 and id = 2")
    assert qep.plan["Node Type"] == "Result"
    assert qep.root[0].plan["Node Type"] == "Index Scan"
    assert qep.root[0].plan["Relation Name"] == "stories"
    assert qep.root[0].plan["Alias"] == "stories"
    assert qep.root[0].plan["Actual Rows"] == 0

    qep = parser("select * from users where id = 1 and id = 2")
    assert qep.plan["Node Type"] == "Result"
    assert qep.root[0].plan["Node Type"] == "Index Scan"
    assert qep.root[0].plan["Relation Name"] == "users"
    assert qep.root[0].plan["Alias"] == "users"
    assert qep.root[0].plan["Actual Rows"] == 0

    qep = parser("select * from comments where id = 1 and id = 2")
    assert qep.plan["Node Type"] == "Result"
    assert qep.root[0].plan["Node Type"] == "Index Scan"
    assert qep.root[0].plan["Relation Name"] == "comments"
    assert qep.root[0].plan["Alias"] == "comments"
    assert qep.root[0].plan["Actual Rows"] == 0

    qep = parser("select * from stories where id = 1 or id = 2")
    assert qep.plan["Node Type"] == "Bitmap Heap Scan"
    assert qep.plan["Relation Name"] == "stories"
    assert qep.plan["Alias"] == "stories"
    assert qep.plan["Actual Rows"] == 2

    qep = parser("select * from users where id = 1 or id = 2")
    assert qep.plan["Node Type"] == "Bitmap Heap Scan"
    assert qep.plan["Relation Name"] == "users"
    assert qep.plan["Alias"] == "users"
    assert qep.plan["Actual Rows"] == 2


def test_qep_find(parser: qepparser.QEPParser):
    """Test that the QEP find method works as expected."""

    qep = parser("select * from stories")
    assert qep.root.findval("Node Type", "Seq Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "stories") == [qep.plan]
    assert qep.root.findval("Alias", "stories") == [qep.plan]
    assert qep.root.findval("Actual Rows", 2) == [qep.plan]

    qep = parser("select * from users")
    assert qep.root.findval("Node Type", "Seq Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "users") == [qep.plan]
    assert qep.root.findval("Alias", "users") == [qep.plan]
    assert qep.root.findval("Actual Rows", 2) == [qep.plan]

    qep = parser("select * from comments")
    assert qep.root.findval("Node Type", "Seq Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "comments") == [qep.plan]
    assert qep.root.findval("Alias", "comments") == [qep.plan]
    assert qep.root.findval("Actual Rows", 4) == [qep.plan]

    qep = parser("select * from stories where id = 1")
    assert qep.root.findval("Node Type", "Index Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "stories") == [qep.plan]
    assert qep.root.findval("Alias", "stories") == [qep.plan]
    assert qep.root.findval("Actual Rows", 1) == [qep.plan]

    qep = parser("select * from users where id = 1")
    assert qep.root.findval("Node Type", "Index Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "users") == [qep.plan]
    assert qep.root.findval("Alias", "users") == [qep.plan]
    assert qep.root.findval("Actual Rows", 1) == [qep.plan]

    qep = parser("select * from comments where id = 1")
    assert qep.root.findval("Node Type", "Index Scan") == [qep.plan]
    assert qep.root.findval("Relation Name", "comments") == [qep.plan]
    assert qep.root.findval("Alias", "comments") == [qep.plan]
    assert qep.root.findval("Actual Rows", 1) == [qep.plan]


def test_gep_rfind(parser: qepparser.QEPParser):
    """Test that the QEP rfind (recursive find) method works as expected."""

    qep = parser("select * from users where id = 1 or id = 2")
    assert len(qep.root.rfindval("Node Type", "Bitmap Heap Scan")) == 1
    assert len(qep.root.rfindval("Node Type", "BitmapOr")) == 2
    assert len(qep.root.rfindval("Node Type", "Bitmap Index Scan")) == 4

    qep = parser("select * from stories where id = 1 or id = 2")
    assert len(qep.root.rfindval("Node Type", "Bitmap Heap Scan")) == 1
    assert len(qep.root.rfindval("Node Type", "BitmapOr")) == 2
    assert len(qep.root.rfindval("Node Type", "Bitmap Index Scan")) == 4

    qep = parser("select * from comments where id = 1 or id = 2")
    assert len(qep.root.rfindval("Node Type", "Bitmap Heap Scan")) == 1
    assert len(qep.root.rfindval("Node Type", "BitmapOr")) == 2
    assert len(qep.root.rfindval("Node Type", "Bitmap Index Scan")) == 4



from typing import Optional

import sqlglot.expressions as exp

from .qepparser import QEPAnalysis
from .errfmt import ErrorFormatter


class EqWildcardChecker:
    def __init__(self, parsed_sql: exp.Expression, qep_analysis: QEPAnalysis):
        self.parsed_sql = parsed_sql
        self.qep_analysis = qep_analysis

    def check(self) -> Optional[str]:
        """
        Returns warning message if the sql has equals operation to a string with
        wild card character (the '%' character), otherwise None.
        """

        eqs = self.parsed_sql.find_all(exp.EQ)

        def is_wildcard_string_eq(eq):
            return self._is_wildcard_string_literal(
                eq.left
            ) or self._is_wildcard_string_literal(eq.right)

        has_eq_wildcard = any(filter(is_wildcard_string_eq, eqs))

        if not has_eq_wildcard:
            return None

        warning = "Possible use of '=' instead of LIKE for wildcard pattern"
        warning_name = type(self).__name__.rstrip("Checker")

        formatter = ErrorFormatter(warning, warning_name)
        warning_msg = formatter.format()

        return warning_msg

    def _is_wildcard_string_literal(self, operand: exp.Expression) -> bool:
        if type(operand) == exp.Literal and operand.is_string:
            return operand.this.find("%") != -1
        return False



# Written by Tatu Heikkilä, tatu.heikkila@tuni.fi
# Licensed under MIT.
"""Handle semantic analysis modules."""
from typing import Optional, Type, Any
import psycopg
from sqlglot import exp

from .config_values import ConfigValues
from .sqlparser import SqlParser, Column
from .qepparser import QEPAnalysis, QEPParser

# analysis modules
from .cmp_domain_checker import CmpDomainChecker
from .eq_wildcard_checker import EqWildcardChecker
from .implied_expression_checker import ImpliedExpressionChecker
from .inconsistent_expression_checker import InconsistentExpressionChecker
from .strange_having_checker import StrangeHavingChecker
from .subquery_order_by_checker import SubqueryOrderByChecker
from .subquery_select_checker import SubquerySelectChecker
from .sum_distinct_checker import SumDistinctChecker


class SemanticRouter:
    """Analyze given SQL queries via a plethora of analysis modules."""

    def __init__(
        self,
        pg_host: str,
        pg_port: str,
        pg_user: str,
        pg_pass: str,
        pg_name: str,
        config_values: Optional[ConfigValues]
    ):
        """Initialize Postgres connection with given paramaters."""
        self.pg_host: str = pg_host
        self.pg_port: str = pg_port
        self.pg_user: str = pg_user
        self.pg_pass: str = pg_pass
        self.pg_name: str = pg_name
        self.config_values: Optional[ConfigValues] = config_values

    def run_analysis(
        self,
        sql_query: str
    ) -> str:
        """Run analysis modules on SQL query string and get an insightful \
        message in return.

        Semantic router (some day) implements basic heuristics to avoid
        running all the modules on all queries. For now, it is dumb brute
        force router.
        :param sql_query: is a single well-formed query to run analytics on.
        :returns: an insightful message that might include vt100-compatible \
        control codes and newlines (without carriage returns).
        """
        try:
            with psycopg.connect(
                "host=" + self.pg_host +
                " port=" + self.pg_port +
                " dbname=" + self.pg_name +
                " user=" + self.pg_user +
                " password=" + self.pg_pass
            ) as conn:
                sql_parser: SqlParser = \
                    SqlParser(conn)
                sanitized_sql: exp.Expression = \
                    sql_parser.parse_one(sql_query)
                qep_analysis: QEPAnalysis = \
                    QEPParser(conn=conn).parse(sql_query)
                analysis_result: Optional[str] = \
                    None

                columns: list[Column] = \
                    sql_parser.get_query_columns(sanitized_sql)

                def is_disabled_in_config(checker_class: Type[Any]) -> bool:
                    if self.config_values is None:
                        return False
                    check_name = checker_class.__name__.rstrip("Checker")
                    return self.config_values.get(check_name) is False

                # Comparing different domains
                if not is_disabled_in_config(CmpDomainChecker):
                    analysis_result = CmpDomainChecker(
                        sanitized_sql,
                        columns
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # ORDER BY in subquery
                if not is_disabled_in_config(SubqueryOrderByChecker):
                    analysis_result = SubqueryOrderByChecker(
                        sanitized_sql,
                        qep_analysis
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # SELECT in subquery
                if not is_disabled_in_config(SubquerySelectChecker):
                    analysis_result = SubquerySelectChecker(
                        sanitized_sql,
                        sql_parser
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # Implied expression
                if not is_disabled_in_config(ImpliedExpressionChecker):
                    analysis_result = ImpliedExpressionChecker(
                        sanitized_sql,
                        sql_query,
                        conn
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # Strange HAVING clause without GROUP BY
                if not is_disabled_in_config(StrangeHavingChecker):
                    analysis_result = StrangeHavingChecker(
                        sanitized_sql,
                        qep_analysis
                    ).check()

                if analysis_result is not None:
                    return analysis_result

                # SUM/AVG(DISTINCT)
                if not is_disabled_in_config(SumDistinctChecker):
                    analysis_result = SumDistinctChecker(
                        sanitized_sql,
                        qep_analysis
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # Wildcards without LIKE
                if not is_disabled_in_config(EqWildcardChecker):
                    analysis_result = EqWildcardChecker(
                        sanitized_sql,
                        qep_analysis
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

                # Inconsistent expression
                if not is_disabled_in_config(InconsistentExpressionChecker):
                    analysis_result = InconsistentExpressionChecker(
                        sanitized_sql,
                        qep_analysis
                    ).check()

                    if analysis_result is not None:
                        return analysis_result

            return ""  # No semantic errors found

        # SQL parser, QEP parser, or an analysis module exploded:
        except Exception:  # Matches only program errors (see flake8 rule E722)
            return ""
